# -*- coding: utf-8 -*-

import base64
import hmac
import random
import string
import time
from hashlib import sha1
import urllib.parse

from . import __version__

quote = lambda s: urllib.parse.quote(s, safe="")


class PeonyHeaders(dict):
    """
        Dynamic headers for Peony

    This is the base class of :class:`OAuth1Headers` and
    :class:`OAuth2Headers`.
    """

    def __init__(self, compression=True, **kwargs):
        """ Add a nice User-Agent """
        self['User-Agent'] = "peony v%s" % __version__
        if compression:
            self['Accept-Encoding'] = "deflate, gzip"

        super().__init__(**kwargs)

    def prepare_request(self, method, url,
                        headers=None,
                        skip_params=False,
                        **kwargs):
        """
        prepare all the arguments for the request

        Parameters
        ----------
        method : str
            HTTP method used by the request
        url : str
            The url to request
        headers : :obj:`dict`, optional
            Additionnal headers
        skip_params : bool
            Don't use the parameters to sign the request

        Returns
        -------
        dict
            Parameters of the request correctly formatted
        """

        if method.lower() == "post":
            key = "data"
        else:
            key = "params"

        if key in kwargs and not skip_params:
            request_params = {key: kwargs.pop(key)}
        else:
            request_params = {}

        request_params.update(dict(method=method.upper(), url=url))

        request_params['headers'] = self.sign(**request_params,
                                              skip_params=skip_params,
                                              headers=headers)

        if headers is not None:
            request_params['headers'].update(headers)

        kwargs.update(request_params)

        return kwargs

    def prepare_headers(self):
        pass

    def sign(self, *args, headers=None, **kwargs):
        if headers is None:
            return self.copy()
        else:
            return self.copy().update(headers)


class OAuth1Headers(PeonyHeaders):
    """
        Dynamic headers implementing OAuth1

    :meth:`sign` is called before each request

    Parameters
    ----------
    consumer_key : str
        Your consumer key
    consumer_secret : str
        Your consumer secret
    access_token : str
        Your access token
    access_token_secret : str
        Your access token secret
    **kwargs
        Other headers
    """

    def __init__(self, consumer_key, consumer_secret,
                 access_token=None, access_token_secret=None, **kwargs):
        """ create the OAuth1 client """
        super().__init__(**kwargs)

        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.access_token = access_token
        self.access_token_secret = access_token_secret

        self.alphabet = string.ascii_letters + string.digits

    def sign(self, method='GET', url=None,
             data=None,
             params=None,
             skip_params=False,
             headers=None,
             **kwargs):
        """ sign, that is, generate the `Authorization` headers """

        headers = super().sign(headers=headers)

        if data:
            if skip_params:
                default = "application/octet-stream"
            else:
                default = "application/x-www-form-urlencoded"

            if not 'Content-Type' in headers:
                headers['Content-Type'] = default

            params = data

        oauth = {
            'oauth_consumer_key': self.consumer_key,
            'oauth_nonce': self.gen_nonce(),
            'oauth_signature_method': 'HMAC-SHA1',
            'oauth_timestamp': str(int(time.time())),
            'oauth_version': '1.0'
        }

        if self.access_token is not None:
            oauth['oauth_token'] = self.access_token

        oauth['oauth_signature'] = self.gen_signature(method=method, url=url,
                                                      params=params,
                                                      skip_params=skip_params,
                                                      oauth=oauth)

        headers['Authorization'] = "OAuth "

        for key, value in sorted(oauth.items(), key=lambda i: i[0]):
            if len(headers['Authorization']) > len("OAuth "):
                headers['Authorization'] += ", "

            headers['Authorization'] += quote(key) + '="' + quote(value) + '"'

        return headers

    def gen_nonce(self):
        return ''.join(random.choice(self.alphabet) for i in range(32))

    def gen_signature(self, method, url, params, skip_params, oauth):
        signature = method.upper() + "&" + quote(url) + "&"

        if skip_params:
            params = oauth
        else:
            params.update(oauth)

        param_string = ""

        for key, value in sorted(params.items(), key=lambda i: i[0]):
            if param_string:
                param_string += "&"

            if key == "q":
                encoded_value = urllib.parse.quote(value, safe="$:!?/()'*@")
                param_string += quote(key) + "=" + encoded_value
            else:
                param_string += quote(key) + "=" + quote(value)

        signature += quote(param_string)

        key = quote(self.consumer_secret).encode() + b"&"
        if self.access_token_secret is not None:
            key += quote(self.access_token_secret).encode()

        signature = hmac.new(key, signature.encode(), sha1)

        signature = base64.b64encode(signature.digest()).decode().rstrip("\n")
        return signature


class OAuth2Headers(PeonyHeaders):
    """
        Dynamic headers implementing OAuth2

    Parameters
    ----------
    consumer_key : str
        Your consumer key
    consumer_secret : str
        Your consumer secret
    client : peony.BasePeonyClient
        The client to authenticate
    bearer_token : :obj:`str`, optional
        Your bearer_token
    **kwargs
        Other headers
    """

    def __init__(self, consumer_key, consumer_secret, client,
                 bearer_token=None, **kwargs):
        super().__init__(**kwargs)
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.client = client
        if bearer_token:
            self.set_token(bearer_token)
        else:
            self.prepare_headers = self.refresh_token

        self.basic_authorization = self.get_basic_authorization()

    def get_basic_authorization(self):
        encoded_keys = map(quote, (self.consumer_key, self.consumer_secret))
        creds = ':'.join(encoded_keys).encode('utf-8')

        auth = "Basic " + base64.b64encode(creds).decode('utf-8')

        return {'Authorization': auth}

    def set_token(self, access_token):
        self['Authorization'] = "Bearer " + access_token

    async def invalidate_token(self, token=None):
        token = token or self.pop('Authorization')[len("Bearer "):]
        request = self.client['api', '', ''].oauth2.invalidate_token.post
        await request(access_token=token, _headers=self.basic_authorization)

    async def refresh_token(self):
        if 'Authorization' in self:
            await self.invalidate_token()

        request = self.client['api', "", ""].oauth2.token.post
        token = await request(grant_type="client_credentials",
                              _headers=self.basic_authorization,
                              _json=True,
                              _is_init_task=True)

        self.set_token(token['access_token'])



# -*- coding: utf-8 -*-

from . import utils
from .commands import Commands
from .tasks import task
from ..utils import print_error, get_args


class EventHandler(task):

    def __init__(self, *args, func, event, prefix=None, **values):
        super().__init__(func)

        self.prefix = prefix
        self.is_event = event

        if prefix is not None:
            self.command = Commands(prefix=prefix)

    def __call__(self, *args):
        argcount = self.__wrapped__.__code__.co_argcount

        if hasattr(self, 'command'):
            args = (*args, self.command,)

        args = args[:argcount]
        return super().__call__(*args)

    def __repr__(self):
        return "<{clsname}: event:{event} prefix:{prefix}>".format(
            clsname=self.__class__.__name__,
            prefix=self.prefix,
            event=self.is_event
        )

    @classmethod
    def event_handler(cls, *args, event, prefix=None, **values):

        def decorator(func):
            event_handler = cls(
                *args,
                func=func,
                event=event,
                prefix=prefix,
                **values
            )

            return event_handler

        return decorator


class EventStream:

    def __init__(self, client):
        self._client = client
        self.functions = [getattr(self, func)
                          for func in dir(self) if self._check(func)]

    def __getitem__(self, key):
        return self._client[key]

    def __getattr__(self, key):
        return getattr(self._client, key)

    @property
    def stream_request(self):
        clsname = self.__class__.__name__
        msg = "You must overload stream_request property in " + clsname
        raise RuntimeError(msg)

    @utils.restart_on(Exception)
    async def start(self):
        if callable(self.stream_request):
            stream_request = self.stream_request()
        else:
            stream_request = self.stream_request

        async with stream_request as resource:
            async for data in resource:
                try:
                    await self._run(data)
                except Exception as e:
                    msg = "error in %s._start:\n" % self.__class__.__name__
                    print_error(msg)

    def _check(self, func):
        if not func.startswith("_"):
            return isinstance(getattr(self, func), EventHandler)
        else:
            return False

    def _get(self, data):
        try:
            for event_handler in self.functions:
                if event_handler.is_event(data):
                    return event_handler

        except:
            msg = "error in %s._get:\n" % self.__class__.__name__
            print_error(msg)

    async def _run(self, data):
        event_handler = self._get(data)

        try:
            if event_handler:
                coro = event_handler(self, data)
                return await utils.execute(coro)

        except:
            fmt = "error occurred while running {classname} {handler}:\n"
            msg = fmt.format(classname=self.__class__.__name__,
                             handler=event_handler.__name__)

            print_error(msg)


def check_setup(func):
    def decorated(self, client):
        if not self.is_setup:
            self.setup(client)

        return func(self, client)

    return decorated


class EventStreams(list):

    def __init__(self):
        super().__init__()
        self.is_setup = False

    @check_setup
    def get_tasks(self, client):
        return [client.loop.create_task(stream.start()) for stream in self]

    @check_setup
    def get_task(self, client):
        if len(self) == 1:
            return client.loop.create_task(self[0].start())
        elif self:
            raise RuntimeError("more than 1 event stream")
        else:
            raise RuntimeError("no event stream")

    def setup(self, client):
        for i in range(len(self)):
            self[i] = self[i](client=client)

        self.is_setup = True



# -*- coding: utf-8 -*-

from functools import wraps

from .event_handlers import EventHandler
from ..utils import get_args


def get_value(func):
    value = func()

    if value is None:
        value = func.__name__

    return value

class Handler:

    def __init__(self, event):
        self.event = event

    def __call__(self, func):
        return EventHandler(func=func, event=self.event)

    def with_prefix(self, prefix):

        def decorated(func):
            return EventHandler(func=func, event=self.event, prefix=prefix)

        return decorated


class Event:

    def __init__(self, func, name):
        self._func = func
        self.handler = Handler(func)
        self.name = name

    def envelope(self):

        def enveloped_event(data):
            return 'for_user' in data and self._func(data.get('message'))

        return self.__class__(enveloped_event, self.name)

    for_user = envelope

    def __str__(self):
        return "Event {name}".format(name=self.name)

    def __repr__(self):
        return str(self)


class Events(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aliases = {}

    def __getattr__(self, key):
        return self[key]

    def __setitem__(self, key, func):
        super().__setitem__(key, Event(func, key))

    def _set_aliases(self, *keys, func):
        for key in keys:
            if "{name}" in key:
                key = key.format(name=func.__name__)

            self[key] = func
            self.aliases[key] = self[func.__name__]

    def alias(self, *keys):

        def decorator(func):
            func = self(func)
            self._set_aliases(*keys, func=func)

            return func

        return decorator

    def event_alias(self, *keys):

        def decorator(func):
            func = self.event(func)
            self._set_aliases(*keys, func=func)

            return func

        return decorator

    def event(self, func):
        if not len(get_args(func)):
            value = get_value(func)

            @wraps(func)
            def decorated(data):
                return data.get('event') == value

            self[func.__name__] = decorated
            self['on_' + func.__name__] = decorated
            return decorated
        else:
            self[func.__name__] = func
            return func

    def __call__(self, func):
        if not len(get_args(func)):
            value = get_value(func)

            @wraps(func)
            def decorated(data):
                return value in data

            self[func.__name__] = decorated
            return decorated
        else:
            self[func.__name__] = func
            return func

    @property
    def no_aliases(self):
        return {key: value for key, value in self.items()
                if key not in self.aliases}


events = Events()

on = 'on_{name}'


@events.alias('on_connect', 'connect')
def friends(data):
    return 'friends' in data or 'friends_str' in data


@events.alias(on, 'on_dm')
def direct_message():
    pass


@events.alias(on)
def tweet(data):
    return 'text' in data and not 'event' in data


@events.alias(on, 'deleted_tweet')
def delete():
    pass


@events.alias('location_deleted')
def scrub_geo():
    pass


@events.event
def limit():
    pass


@events.event
def status_withheld():
    pass


@events.event
def user_withheld():
    pass


@events.alias(on)
def disconnect():
    pass


# warnings

@events.alias(on)
def warning():
    pass


@events.alias(on)
def stall_warning(data):
    return (warning(data) and
            data.get('warning').get('code') == "FALLING_BEHIND")


@events.alias(on)
def too_many_follows(data):
    return (warning(data) and
            data.get('warning').get('code') == "FOLLOWS_OVER_LIMIT")

# events, the data looks like; {"event": EVENT_NAME, ...}

@events.event
def access_revoked():
    pass


@events.event
def follow():
    pass


@events.event
def unfollow():
    pass


@events.event
def access_revoked():
    pass


@events.event
def block():
    pass


@events.event
def unblock():
    pass


@events.event
def favorite():
    pass


@events.event
def unfavorite():
    pass


@events.event
def list_created():
    pass


@events.event
def list_destroyed():
    pass


@events.event
def list_updated():
    pass


@events.event
def list_member_added():
    pass


@events.event
def list_member_removed():
    pass


@events.event
def list_user_subscribed():
    pass


@events.event
def list_user_unsubscribed():
    pass


@events.event
def quoted_tweet():
    pass


@events.event
def user_update():
    pass


# Site stream control messages

@events.alias('control_message')
def control():
    pass


# Internal peony events

@events.alias(on, 'on_restart', 'restart')
def stream_restart():
    pass


@events.alias(on, 'reconnect', 'on_reconnect')
def reconnecting_in():
    pass



# -*- coding: utf-8 -*-

import asyncio
import sys

import aiohttp

from . import exceptions, utils
from .exceptions import StreamLimit
from .general import rate_limit_notices

if int(aiohttp.__version__.split('.')[0]) < 2:
    ClientPayloadError = aiohttp.errors.ContentEncodingError
    ClientConnectionError = aiohttp.errors.ClientDisconnectedError
else:
    ClientPayloadError = aiohttp.ClientPayloadError
    ClientConnectionError = aiohttp.ClientConnectionError

RECONNECTION_TIMEOUT = 5
MAX_RECONNECTION_TIMEOUT = 320
DISCONNECTION_TIMEOUT = 0.25
MAX_DISCONNECTION_TIMEOUT = 16
ENHANCE_YOUR_CALM_TIMEOUT = 60


class StreamResponse:
    """
        Asynchronous iterator for streams

    Parameters
    ----------
    *args : optional
        Positional arguments
    _headers : dict
        Headers to authorize the request
    session : :obj:`aiohttp.Session`, optional
        Session used by the request
    reconnect : :obj:`int`, optional
        Time to wait for on error
    loads : function, optional
        function used to decode the JSON data received
    timeout : :obj:`int`, optional
        Timeout for requests
    _timeout : :obj:`int`, optional
        Stream timeout, the connection will be closed if this timeout
        is exceeded
    _error_handler : function, optional
        Request's error handler
    **kwargs
        Keyword parameters of the request
    """

    def __init__(self, *args,
                 client,
                 session=None,
                 loads=utils.loads,
                 timeout=10,
                 **kwargs):

        self.client = client
        self.session = self.client._session if session is None else session
        self.loads = loads
        self.timeout = timeout
        self.args = args
        self.kwargs = kwargs
        self.reconnecting = False
        self.reconnection = 0
        self.enhance_calm = 0
        self.disconnection = 0

    def connect(self):
        """
            Connect to the stream

        Returns
        -------
        asyncio.coroutine
            The streaming response
        """
        kwargs = self.client.headers.prepare_request(**self.kwargs)
        request = self.client.error_handler(self.session.request)

        if 'proxy' not in kwargs:
            kwargs['proxy'] = self.client.proxy

        return request(*self.args, timeout=0, **kwargs)

    def initialize_timeouts(self):
        self.reconnection = 0
        self.enhance_calm = 0
        self.disconnection = 0

    async def __aiter__(self):
        """
            Create the connection

        Returns
        -------
        self

        Raises
        ------
        exception.PeonyException
            On a response status != 2xx
        """
        with aiohttp.Timeout(self.timeout):
            self.response = await self.connect()

        if self.response.status == 200:
            self.initialize_timeouts()
        elif self.response.status == 500:
            self.increase_disconnetion_timeout()
        elif self.response.status in range(501, 600):
            self.increase_reconnection_timeout()
        elif self.response.status in (420, 429):
            self.increase_enhance_calm_timeout()

            print("Enhance Your Calm response received from Twitter. "
                  "If you didn't restart your program frenetically "
                  "then there is probably something wrong with it. "
                  "Make sure you are not opening too many connections to "
                  "the endpoint you are currently using by checking "
                  "Twitter's Streaming API documentation out: "
                  "https://dev.twitter.com/streaming/overview\n"
                  "The stream will restart in %ss." % self.enhance_calm,
                  file=sys.stderr)
        else:
            raise await exceptions.throw(self.response)

        return self

    async def __anext__(self):
        """
            Decode each line using json

        Returns
        -------
        dict
            Decoded JSON data
        """
        line = b''
        try:
            timeouts = self.disconnection, self.reconnection, self.enhance_calm
            reconnect = max(timeouts)

            if self.reconnecting is not False:
                return await self.restart_stream()
            elif reconnect:
                return await self.initialize_restart(reconnect=reconnect)

            while not line:
                with aiohttp.Timeout(90):
                    line = await self.response.content.readline()
                    line = line.rstrip(b'\r\n')

            if line in rate_limit_notices:
                raise StreamLimit(line)

            return self.loads(line)

        except asyncio.TimeoutError:
            return await self.initialize_restart()

        except ClientPayloadError:
            return await self.initialize_restart()

        except ClientConnectionError:
            self.increase_disconnetion_timeout()
            return await self.initialize_restart(reconnect=self.disconnection)

        except Exception as e:
            return await self.initialize_restart(error=e)

    def increase_disconnetion_timeout(self):
        if not self.disconnection >= MAX_DISCONNECTION_TIMEOUT:
            self.disconnection += DISCONNECTION_TIMEOUT

    def increase_reconnection_timeout(self):
        if self.reconnection == 0:
            self.reconnection = RECONNECTION_TIMEOUT
        elif not self.reconnection >= MAX_RECONNECTION_TIMEOUT:
            self.reconnection *= 2

    def increase_enhance_calm_timeout(self):
        if self.enhance_calm == 0:
            self.enhance_calm = ENHANCE_YOUR_CALM_TIMEOUT
        else:
            self.enhance_calm *= 2

    async def initialize_restart(self, reconnect=0, error=None):
        """
            Restart the stream on error

        Parameters
        ----------
        reconnect : :obj:`int`, optional
            Time to wait for before reconnecting
        error : :class:`Exception`, optional
            Whether to print the error or not
        """
        if reconnect is not None:
            if error:
                utils.print_error()

            self.reconnecting = reconnect
            return {
                'reconnecting_in': reconnect,
                'error': error
            }
        else:
            if error is not None:
                raise error

    async def restart_stream(self):
        """
            Restart the stream on error
        """
        await self.response.release()
        await asyncio.sleep(self.reconnecting)
        await self.__aiter__()
        self.reconnecting = False
        return {'stream_restart': True}


class StreamContext:
    """
        A context that should close the request on exit

    Parameters
    ----------
    method : str
        HTTP method used to make the request
    url : str
        The API endpoint
    *args : optional
        Positional arguments
    **kwargs
        Keyword parameters of the request and of :class:`StreamResponse`
    """

    def __init__(self, method, url, client, *args, **kwargs):
        self.method = method
        self.url = url
        self.client = client
        self.args = args
        self.kwargs = kwargs

    async def __aenter__(self):
        """
            Create stream

        Returns
        -------
        StreamResponse
            The stream iterator
        """
        await self.client.setup()
        self.stream = StreamResponse(method=self.method, url=self.url,
                                     *self.args, client=self.client,
                                     **self.kwargs)

        return self.stream

    async def __aexit__(self, *args, **kwargs):
        """
            Close the response on error
        """

        if hasattr(self.stream, "response"):
            self.stream.response.close()
